/* David Leonard, 2002. Public domain. */
/* $Id: main.c,v 1.4 2002/07/30 07:33:24 d Exp $ */

#include <netdb.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <err.h>
#include <pcap.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <time.h>
#include <fcntl.h>
#include <net/ppp_defs.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>

struct arphdr { int ignore; };
struct ifnet { int ignore; };
#include <sys/queue.h>
#include <netinet/if_ether.h>

#include "proto.h"

/* tcpdump reader - view TCP streams and possibly decode their contents */

static const char version[] = "1.1";

/* Connection key (src=server dst=client) */
struct tcpkey {
	struct in_addr	ip_src, ip_dst;
	u_int16_t	th_sport, th_dport;
};

/* Connection state */
struct tcpstate {
	struct tcpstate *next;
	struct tcpkey	key;
	int	connid;
	int	state;
#define	CS	1		/* client syn seen */
#define	SS	2		/* server syn seen */
#define CF	4		/* client fin seen */
#define SF	8		/* server fin seen */
#define STATE_ESTABLISHED	(CS|SS)
#define STATE_CLOSED		(CS|SS|CF|SF)
#define STATE_UNOPENED		(0)
	u_int32_t	sseq, dseq;	/* next expected sequence numbers */
	struct protostate	*proto;
};

static int next_connid = 1;			/* next connection ID */
static struct tcpstate *tcpstates = NULL;	/* connection list */

/* Return true if tcpstate keys match (ie same connection) */
static int
matchstate(t1, t2)
	struct tcpstate *t1, *t2;
{
	return t1->key.ip_src.s_addr == t2->key.ip_src.s_addr
	    && t1->key.ip_dst.s_addr == t2->key.ip_dst.s_addr
	    && t1->key.th_sport == t2->key.th_sport
	    && t1->key.th_dport == t2->key.th_dport;
}

/* Search for existing tcp state */
static int
state_find(ip, tcp, state, reverse)
	struct ip *ip;
	struct tcphdr *tcp;
	struct tcpstate *state;
	int reverse;
{
	struct tcpstate *t;

	/* Set up a state key based on ip and tcp headers */
	state->key.ip_src = reverse ? ip->ip_dst : ip->ip_src;
	state->key.ip_dst = reverse ? ip->ip_src : ip->ip_dst;
	state->key.th_sport = reverse ? tcp->th_dport : tcp->th_sport;
	state->key.th_dport = reverse ? tcp->th_sport : tcp->th_dport;

	/* Look for existing state - not very efficient :( */
	for (t = tcpstates; t; t = t->next) {
		if (matchstate(t, state)) {
			memcpy(state, t, sizeof *state);
			return 1;
		}
	}

	return 0;
}

/* Store a modified tcp state back into the list */
static void
state_set(state)
	struct tcpstate *state;
{
	struct tcpstate *t;

	for (t = tcpstates; t; t = t->next)
		if (matchstate(t, state)) {
			struct tcpstate *next = t->next;
			memcpy(t, state, sizeof *t);
			t->next = next;
			return;
		}
	/* Insert into top */
	t = (struct tcpstate *)malloc(sizeof *t);
	memcpy(t, state, sizeof *t);
	t->next = tcpstates;
	tcpstates = t;
}

/* Remove a tcp state from the list */
static void
state_del(ip, state)
	struct tcpstate *state;
{
	struct tcpstate  **tp;

	for (tp = &tcpstates; *tp; tp = &((*tp)->next))
		if (matchstate(*tp, state)) {
			struct tcpstate *t = *tp;
			*tp = t->next;
			free(t);
			return;
		}
}

/* Convert IP address to a short name */
static char *
iptoname(addr)
	struct in_addr addr;
{
	struct hostent *he;
	static char name[256];
	char *p;

	/* eek should use /etc/hosts */
	if (addr.s_addr == inet_addr("10.100.253.1"))
		return "thingy";

	he = gethostbyaddr((char *)&addr, sizeof addr, AF_INET);
	if (!he)
		return inet_ntoa(addr);
	strncpy(name, he->h_name, sizeof name);
	p = strchr(name, '.');
	if (p) *p = '\0';
	return name;
}

/* Process a packet for the datalink layer, ignore non-TCP packets */
static void
datalink(context, hdr, data)
	u_char *context;
	const struct pcap_pkthdr *hdr;
	const u_char *data;
{
	struct tm tm;
	char tmbuf[256];
	int datalink_type = (int)context;
	const u_char *p, *pend;
	const struct ip *ip;
	const struct tcphdr *tcp;
	struct tcpstate state;
	int datalen, reverse, f, oldstate, freshstate;
	const char *op = NULL;
	u_int32_t *seq;

	p = data;

	/* Datalink: ignore non-IP */
	if (datalink_type == DLT_PPP) {
		if (PPP_ADDRESS(p) == PPP_ALLSTATIONS
		    && PPP_PROTOCOL(p) == PPP_IP)
			p += PPP_HDRLEN;
		else
			return;	/* ignore non-IP packets */
	} else if (datalink_type == DLT_EN10MB) {
		struct ether_header eh;

		memcpy(&eh, p, sizeof eh);	/* avoid bus alignment probs */
		if (ntohs(eh.ether_type) == ETHERTYPE_IP)
			p += ETHER_HDR_LEN;
		else
			return; /* ignore non-IP packets */
	} else {
		errx(1, "unsupported datalink type %d", datalink_type);
	}

	/* IP: ignore non-TCPv4 */
	ip = (struct ip *)p;
	if (ip->ip_v != IPVERSION || ip->ip_p != IPPROTO_TCP)
		return;
	pend = p + ntohs(ip->ip_len);
	p += ip->ip_hl << 2;

	/* TCP */
	tcp = (struct tcphdr *)p;

	/* Skip to data */
	p += tcp->th_off << 2;

	/* Find/create connection state */
	memset(&state, 0, sizeof state);
	if (state_find(ip, tcp, &state, 0)) {
		reverse = 0;			/* src->dst = server->client */
	} else if (state_find(ip, tcp, &state, 1)) {
		reverse = 1;			/* src->dst = client->server */
	} else {
		/* assume first packet we see is client->server */
		reverse = 1;
		state.state = STATE_UNOPENED;
		state.connid = 0;
		state.proto = NULL;
	}

	/* 
	 * Assume that packets we see being sent are delivered
	 */
	f = tcp->th_flags;
#define A	TH_ACK					/* convenience */
#define S	TH_SYN
#define F	TH_FIN
#define R	TH_RST

	oldstate = state.state;
	op = NULL;
	if (reverse) {	/* client->server */
		if (f & S) { state.state |= CS; op = "connect"; }
		if (f & F) {
			state.state |= CF; 
			if (state.state & SF)
				op = "closed";
			else
				op = "close (by client)";
		}
		seq = &state.dseq;
	} else {	/* server -> client */
		if ((f & R) /* && (state.state == CS || state.state == SS) */) {
			state.state = STATE_UNOPENED;
			op = "reset (reject)";
		}
		if (f & S) { state.state |= SS; op = "accept"; }
		if (f & F) {
			state.state |= SF;
			if (state.state & CF)
				op = "closed";
			else
				op = "close (by server)";
		}
		seq = &state.sseq;
	}

	if (f & S)
		*seq = ntohl(tcp->th_seq) + 1;

	/* Give conn IDs to new connections */
	freshstate = (state.connid == 0);
	if (oldstate != state.state && state.state == STATE_ESTABLISHED)  {
		state.connid = next_connid++;
		state.proto = proto_open(state.key.th_sport, state.connid);
	}

	/* Print the flow ID */
	printf("%-3d", state.connid);

	/* Print the packet time */
	if (localtime_r(&hdr->ts.tv_sec, &tm) == NULL)
		errx(1, "localtime_r");
	if (strftime(tmbuf, sizeof tmbuf, "%T", &tm) == 0)
		errx(1, "strftime");
	printf(" %s.%06u", tmbuf, hdr->ts.tv_usec);

	/* Print the IP and TCP addresses if this is newly constructed state */
	if (freshstate) {
	    if (reverse) {
	        printf(" %s:%u", iptoname(ip->ip_src), ntohs(tcp->th_sport));
	        printf(" >");
	        printf(" %s:%u", iptoname(ip->ip_dst), ntohs(tcp->th_dport));
	    } else {
	        printf(" %s:%u", iptoname(ip->ip_dst), ntohs(tcp->th_dport));
	        printf(" <");
	        printf(" %s:%u", iptoname(ip->ip_src), ntohs(tcp->th_sport));
	    }
	} 
	printf(" %c", reverse ? 'C' : 'S');

	if (f & TH_URG) printf(" urgent");

	datalen = pend - p;
	if (op == NULL)
		op = datalen ? "data" : "ack";
	printf(" %s", op);
	if (datalen) printf(" %d (0x%x) bytes", datalen, datalen);

	/* Forget connections that just closed, or never opened */
	if (oldstate != state.state && 
	    (state.state == STATE_CLOSED || state.state == STATE_UNOPENED))
	{
		if (state.proto)
			proto_close(state.proto);
		state_del(&state);
		printf("\n");
		return;
	}

	/* Check for duplicates packets (ie out of the expected window) */
	if (datalen) {
		/* printf(" seq %08x\n", ntohl(tcp->th_seq)); */
		if (ntohl(tcp->th_seq) != *seq) {
			/* XXX out-of-order packets not implemented */
			printf(" dup\n");
		} else {
			*seq += datalen;
			printf("\n");
			proto_feed(state.proto, reverse, p, datalen);
		}
	} else
		printf("\n");

	/* Save any changes made to the state */
	state_set(&state);
}

/* main */
int
main(argc, argv)
	int argc;
	char *argv[];
{
	int ch;
	extern int optind;
	extern char *optarg;
	char *rfile = NULL;
	pcap_t *p;
	char errbuf[PCAP_ERRBUF_SIZE];
	int error = 0;
	int datalink_type;

	/* Process command line options */
	while ((ch = getopt(argc, argv, "r:")) != -1)
		switch (ch) {
		case 'r':
			rfile = optarg;
			break;
		default:
			error = 1;
		}
	if (argc != optind)
		error = 1;
	if (rfile == NULL)
		error = 1;
	if (error) {
		fprintf(stderr, "tcpview version %s\n", version);
		fprintf(stderr, "usage: %s -r tcpdump-file\n",
		    argv[0]);
		exit(1);
	}

	/* Open captured file */
	p = pcap_open_offline(rfile, errbuf);
	if (!p) 
		errx(1, "%s: %s", rfile, errbuf);
	datalink_type = pcap_datalink(p);

	/* Dump the packets */
	if (pcap_loop(p, -1, datalink, (u_char *)datalink_type) == -1)
		errx(1, "%s", pcap_geterr(p));

	/* Clean up */
	pcap_close(p);
	exit(0);
}

