/* David Leonard, 2002. Public domain. */
/* $Id: proto.c,v 1.2 2002/07/30 07:33:03 d Exp $ */

#include <stdio.h>
#include <err.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include "proto.h"

/* Protocol decoder utility routines */

extern struct proto *protos[], *default_proto;

/* State associated with each protocol decoder */
struct protostate {
	struct proto *proto;	/* dispatch table */
	void *state;		/* private decoder state */
	int connid;
	struct buf {		/* per-direction buffer */
		unsigned char *data;
		int len;
		unsigned long offset;
	} buf[2];
};

struct context {		/* printing context */
	struct protostate *ps;
	int	reverse;
	int	off;
	struct buf *buf;
};

/* Find a proto decoder and initialise it */
struct protostate *
proto_open(port, connid)
	int port;
	int connid;
{
	struct protostate *ps = (struct protostate *)malloc(sizeof *ps);
	struct proto **p;

	ps->proto = NULL;
	for (p = protos; *p; p++)
		if ((*p)->port == port) 
			break;
	if (*p)
		ps->proto = *p;
	else
		ps->proto = default_proto;
	ps->connid = connid;
	ps->state = ps->proto->statesize ? malloc(ps->proto->statesize) : NULL;
	ps->buf[0].data = ps->buf[1].data = NULL;
	ps->buf[0].len = ps->buf[1].len = 0;
	ps->buf[0].offset = ps->buf[1].offset = 0;
	if (ps->proto->init)
		(*ps->proto->init)(ps->state);

	return ps;
}

/* Close a protocol decoder */
void
proto_close(ps)
	struct protostate *ps;
{

	if (ps->proto->final)
		(*ps->proto->final)(ps->state, ps->buf[0].len, ps->buf[1].len);
	if (ps->state)
		free(ps->state);
	if (ps->buf[0].data)
		free(ps->buf[0].data);
	if (ps->buf[1].data)
		free(ps->buf[1].data);
	free(ps);
}

/* Called when new data is ready for a protocol decoder. */
void
proto_feed(ps, reverse, data, datalen)
	struct protostate *ps;
	int reverse;
	const unsigned char *data;
	int datalen;
{
	struct buf *buf = &ps->buf[reverse];
	struct context context;
	int oldoff;

	if (datalen == 0)
		return;

	/* Append new data to the decoder buffer */
	if (buf->data) {
		buf->data = realloc(buf->data, buf->len + datalen);
		if (buf->data == NULL)
			errx(1, "realloc");
		memcpy(buf->data + buf->len, data, datalen);
		buf->len += datalen;
	} else {
		buf->data = malloc(datalen);
		memcpy(buf->data, data, datalen);
		buf->len = datalen;
	}

	/* Set up a printing context */
	context.ps = ps;
	context.reverse = reverse;
	context.off = 0;		/* offset into buf.data */
	context.buf = buf;

	/* Keep calling the decoder until it won't print any more */
	oldoff = 0;
	do { 
		oldoff = context.off;
		(*ps->proto->decode)(ps->state, &context, reverse,
		    buf->data + context.off, buf->len - context.off);
		if (context.off > buf->len)
			errx(1, "internal error: decode printed too much");
	} while (context.off != oldoff && context.off < buf->len);

	/* Free the part of the buffer just consumed */
	buf->offset += context.off;
	if (context.off == 0) {
		/* nothing */
	} else if (context.off == buf->len) {
		free(buf->data);
		buf->data = NULL;
		buf->len = 0;
	} else {
		unsigned char *d = malloc(buf->len - context.off);
		int dlen = buf->len - context.off;
		memcpy(d, buf->data + context.off, dlen);
		free(buf->data);
		buf->data = d;
		buf->len = dlen;
	}
}

/* Helper function to print decoded information */
void
proto_print(void *ctxt, const unsigned char *end, const char *fmt, ...)
{
	struct context *context = (struct context *)ctxt;
	struct protostate *ps = context->ps;
	int i;
	unsigned char *data = context->buf->data;
	va_list ap;
	int firstline = 1;

	if (end < data || end > data + context->buf->len) {
		warnx("proto_print: end pointer is outside buffer (\"%s\")", fmt);
		abort();
	}

	if (fmt == NULL || fmt[0] == '\0')
		firstline = 0;

	while (firstline || (data + context->off < end)) {
		printf("%-3d %c %04lx:", 
			ps->connid, 
			context->reverse ? 'C' : 'S',
			context->buf->offset + context->off);
		for (i = 0; i < 16; i++) {
			if (i % 4 == 0)
				printf(" ");
			if (data + context->off < end) {
				printf("%02x", data[context->off]);
				context->off++;
			} else
				printf("  ");
		}

		if (firstline) {
			printf(" ");
			va_start(ap, fmt);
			vfprintf(stdout, fmt, ap);
			va_end(ap);
			firstline = 0;
		}
		printf("\n");
	}
}

/* Quote a string in a sensible way */
const char *
quotestr(s, len)
	const char *s;
	int len;
{
	static char buf[256];
	char *p, *end;
	int last0 = 0;

	end = buf + sizeof buf;
	p = buf;

	*p++ = '\"';
	for (; len > 0 && p < end - 10; s++, len--) {
		if (*s == 0) {
			if (!last0) {
				*p++ = '\\';
				*p++ = '0';
				last0 = 1;
				continue;
			}
			last0++;
			continue;
		}
		if (last0) {
			if (last0 > 1) {
				if (last0 > 2) {
					*p++ = '.';
					*p++ = '.';
					*p++ = '.';
				}
				*p++ = '\\';
				*p++ = '0';
			}
			last0 = 0;
		}
		switch (*s) {
		case '\"':
		case '\'':
		case '\\':
			*p++ = '\\';
			*p++ = *s;
			break;
		case '\n':
			*p++ = '\\'; *p++ = 'n'; break;
		case '\r':
			*p++ = '\\'; *p++ = 'r'; break;
		case '\t':
			*p++ = '\\'; *p++ = 't'; break;
		default:
			if (*s >= ' ' && *s <= '~')
				*p++ = *s;
			else {
				unsigned char v = *(unsigned char *)s;
				*p++ = '\\';
				*p++ = 'x';
				*p++ = "0123456789abcdef"[v / 16];
				*p++ = "0123456789abcdef"[v % 16];
			}
		}
	}
	*p++ = '\"';
	if (last0 > 1 || p >= end - 10 + 1) {
		*p++ = '.';
		*p++ = '.';
		*p++ = '.';
	}
	*p++ = '\0';
	return buf;
}
