/* David Leonard, 2002. Public domain. */
/* $Id: proto_cerdisp.c,v 1.1.1.1 2002/01/29 04:48:28 d Exp $ */

#include <stdio.h>
#include <err.h>
#include <sys/types.h>
#include <machine/endian.h>

/* 
 * Protocol decoder for "Remote Control Display for Windows CE"
 */

#include "proto.h"

struct state {
	int	val;
	u_int32_t prelen;
	long	bytes;
#define PREAMBLELEN	0
#define PREAMBLE	1
#define FRAMEHEADER	2
#define FRAME		3
};

/* Keycode table */
static char *keytab[256] = {
	/*0  1   2   3    4   5   6   7    8   9   a   b    c   d   e   f*/
/*00*/	000,000,000,000, 000,000,000,000, "Backspace","Tab",000,000,
					                   000,"Return",000,000,
/*10*/	"Shift","Control",000,000,
			 "Capslock",000,000,000,
					  000,000,000,"Esc",
							   000,000,000,000,
/*20*/	"Space","PgUp","PgDn","End",
			 "Home","Left","Up","Right",
					 "Down",000,000,000,
							   000,"Ins","Del",000,
/*30*/	"0","1","2","3", "4","5","6","7", "8","9",000,000, 000,000,000,000,
/*40*/	000,"a","b","c", "d","e","f","g", "h","i","j","k", "l","m","n","o", 
/*50*/	"p","q","r","s", "t","u","v","w", "x","y","z",000, 000,000,000,000,
/*60*/	"k0","k1","k2","k3",
			 "k4","k5","k6","k7",
					  "k8","k9","Mul","Add",
							   000,"Sub",000,"Div",
/*70*/	"F1","F2","F3","F4",
			 "F5","F6","F7","F8",
					  "F9",000,000,000,
							   000,000,000,000,
/*80*/	000,000,000,000, 000,000,000,000, 000,000,000,000, 000,000,000,000,
/*90*/	000,000,000,000, 000,000,000,000, 000,000,000,000, 000,000,000,000,
/*a0*/	000,000,000,000, 000,000,000,000, 000,000,000,000, 000,000,000,000,
/*b0*/	000,000,000,000, 000,000,000,000, 000,000,";","=", ",","-",".","/",
/*c0*/	"`",000,000,000, 000,000,000,000, 000,000,000,000, 000,000,000,000,
/*d0*/	000,000,000,000, 000,000,000,000, 000,000,000,"[", "\\","]","'","Sleep",
/*e0*/	000,000,000,000, 000,000,000,000, 000,000,000,000, 000,000,000,000,
/*f0*/	000,000,000,000, 000,000,000,000, 000,000,000,000, 000,000,000,000
};

static void
init(void *arg)
{
	struct state *state = (struct state *)arg;
	
	state->val = PREAMBLELEN;
}

static u_int32_t
u32(const unsigned char *d)
{
	u_int32_t v;
	memcpy(&v, d, sizeof v);
	return letoh32(v);
}

static u_int16_t
u16(const unsigned char *d)
{
	u_int16_t v;
	memcpy(&v, d, sizeof v);
	return letoh16(v);
}

static void
decode(arg, context, reverse, data, datalen)
	void *arg;
	void *context;
	int reverse;
	const unsigned char *data;
	int datalen;
{
	struct state *state = (struct state *)arg;
	struct {
		u_int32_t       magic;
		u_int32_t       width;
		u_int32_t       height;
		u_int32_t       maxval;
		u_int32_t       unk1;
		u_int32_t       unk2;
		char            platform[128];
		char            devtype[128];
		char            name[16];
		u_int32_t       unk3;
	} preamble;
	struct {
		u_int32_t       unk0;
		u_int32_t       datalen;
		u_int32_t       unk1;
		u_int32_t       unk2;
		u_int32_t       width;
		u_int32_t       height;
	} framehdr;

	if (!reverse) {
		if (datalen < 4)
			return;
		if ((u16(data) & 0x0100) == 0) {
			/* mouse */
			if (datalen < 6)
				return;
			proto_print(context, data + 6, "mouse %sx=%d y=%d", 
				u16(data) == 2 ? "down " :
				u16(data) == 4 ? "up " : "",
				u16(data + 2),
				u16(data + 4));
			return;
		} else {
			char *key = keytab[u16(data + 2) & 0xff];
			/* key */
			proto_print(context, data + 4, "key %s 0x%02x %s", 
				u16(data) == 0x0100 ? "down" : "up",
				u16(data + 2) & 0xff,
				key ? key : "<unknown>");
			return;
		}
	}

	switch (state->val) {
	case PREAMBLELEN:
		if (datalen < sizeof (u_int32_t))
			return;
		state->prelen = u32(data);
		data += sizeof (u_int32_t);
		proto_print(context, data, "<preamble len 0x%x>",
		    state->prelen);
		state->val = PREAMBLE;
		return;
	case PREAMBLE:
		if (datalen < sizeof preamble)
			return;
		memcpy (&preamble, data, sizeof preamble);

		data += sizeof preamble.magic;
		proto_print(context, data, "magic %04x", 
			letoh32(preamble.magic));

		data += sizeof preamble.width;
		proto_print(context, data, "width %d", 
			letoh32(preamble.width));

		data += sizeof preamble.height;
		proto_print(context, data, "height %d", 
			letoh32(preamble.height));

		data += sizeof preamble.maxval;
		proto_print(context, data, "maxval %d", 
			letoh32(preamble.maxval));

		data += sizeof preamble.unk1 + sizeof preamble.unk2;
		proto_print(context, data, "");

		data += sizeof preamble.platform;
		proto_print(context, data, "platform: \"%.*s\"",
			sizeof preamble.platform, preamble.platform);

		data += sizeof preamble.devtype;
		proto_print(context, data, "devtype: \"%.*s\"",
			sizeof preamble.devtype, preamble.devtype);

		data += sizeof preamble.name;
		proto_print(context, data, "name: \"%.*s\"",
			sizeof preamble.name, preamble.name);

		data += sizeof preamble.unk3;
		proto_print(context, data, "");

		state->val = FRAMEHEADER;
		return;
	case FRAMEHEADER:
		if (datalen < sizeof framehdr)
			return;
		memcpy (&framehdr, data, sizeof framehdr);

		data += sizeof framehdr.unk0;
		proto_print(context, data, "<frame header>");

		data += sizeof framehdr.datalen;
		proto_print(context, data, "<datalen: %d words>",
			letoh32(framehdr.datalen));

		data += sizeof framehdr.unk1 + sizeof framehdr.unk2;
		proto_print(context, data, "(offsets?)");

		data += sizeof framehdr.width;
		proto_print(context, data, "frame width %d",
			letoh32(framehdr.width));

		data += sizeof framehdr.height;
		proto_print(context, data, "frame height %d",
			letoh32(framehdr.height));

		state->val = FRAME;
		state->bytes = letoh32(framehdr.datalen) * 2;
		return;
	case FRAME:
		if (datalen < state->bytes) {
			proto_print(context, data + datalen, "");
			state->bytes -= datalen;
		} else {
			proto_print(context, data + state->bytes, "");
			state->bytes = 0;
			state->val = FRAMEHEADER;
		}
		return;
	default:
		errx(1, "cerdisp: bad decode state");
	}
}

struct proto proto_cerdisp = {
	987,			/* port */
	sizeof (struct state), 	/* state */
	init,			/* init */
	decode,			/* decode */
	NULL			/* final */
};
